from enum import Enum

class TerraformBasicFileName(Enum):
    VARIABLES_TF = "variables.tf"
    OUTPUTS_TF   = "outputs.tf"
    VERSION_TF   = "versions.tf"
    README_MD    = "README.MD"
    README_md    = "README.md"


from typing import Dict


def read_yaml(file_path: str) -> Dict:
    import yaml

    with open(file_path, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return {}

def get_tf_files_paths(folder_path: str):
    from os.path import isfile, join
    import glob

    # List .tf files in the folder
    tf_globs = glob.glob(join(folder_path,"*.tf"))
    # Keep files only
    tf_files = [f for f in tf_globs if isfile(f)]
    return tf_files


from terracheck.types import TerraformObjectType

class CheckerElement():

    def __init__(self, name: str, path: str, terraform_object_type: TerraformObjectType, terraform_resource_type: str, additional_info={}) -> None:
        self.name                    = name
        self.path                    = path
        self.filename                = self.path.split('/')[-1]
        self.terraform_object_type   = terraform_object_type
        self.additional_info         = additional_info
        self.terraform_resource_type = terraform_resource_type

    def to_dict(self):
        return {
            'name': self.name,
            'path': self.path,
            'filename': self.filename,
            'terraform_object_type': self.terraform_object_type.value,
            'additional_info': self.additional_info,
            'terraform_resource_type': self.terraform_resource_type,
        }


import os
from uuid import uuid4

from terracheck.utils.utils import read_yaml, get_tf_files_paths
from terracheck.types.CheckerIssue import CheckerIssue
from terracheck.types.RuleCategory import RuleCategory
from terracheck.types.TerraformBasicFileName import TerraformBasicFileName

class StructureChecker():

    def __init__(self, yaml_config_file_path: str) -> None:
        self.report = {}
        self.config = read_yaml(yaml_config_file_path)
        self.metadata = {}
        self._init_metadata()

    def _init_metadata(self):
        self.metadata["required_files_total_count"] = len(self.config["mandatory_files"])
        self.metadata["required_files_missing_count"] = 0

    def _check_files_presence(self, hcl_folder_path: str):
        for file_type, available_denominations in self.config["mandatory_files"].items():
            is_present = True
            for filename in available_denominations:
                is_present = os.path.isfile(os.path.join(hcl_folder_path, filename))
                if is_present:
                    break
            if not is_present:
                self.report[uuid4()] = CheckerIssue(
                    RuleCategory.MISSING_FILE,
                    f"File '{file_type}' is missing in your terraform folder. Available denominations : {available_denominations}."
                )
                self.metadata["required_files_missing_count"] += 1

    def _check_filenames_nomenclature(self, hcl_folder_path: str):
        if "filenames_nomenclature" in self.config:
            all_tf_files_paths = get_tf_files_paths(hcl_folder_path)
            all_tf_files_names = [os.path.basename(path) for path in all_tf_files_paths]
            files_to_check = list(set(all_tf_files_names)- set([tbfn.value for tbfn in TerraformBasicFileName]))
            for filename in files_to_check:
                startswith_constraint = self.config["filenames_nomenclature"].get("startswith", "")
                if not filename.replace('.tf', '').startswith(startswith_constraint):
                    self.report[uuid4()] = CheckerIssue(
                        RuleCategory.FILENAME_CONSTRAINT,
                        f"File '{filename}' does not respect constraint 'startswith {startswith_constraint}'."
                    ) 
                endswith_constraint = self.config["filenames_nomenclature"].get("endswith", "")
                if not filename.replace('.tf', '').endswith(startswith_constraint):
                    self.report[uuid4()] = CheckerIssue(
                        RuleCategory.FILENAME_CONSTRAINT,
                        f"File '{filename}' does not respect constraint 'endswith {endswith_constraint}'."
                    ) 
                contains_constraint = self.config["filenames_nomenclature"].get("contains", "")
                if not contains_constraint in filename.replace('.tf', ''):
                    self.report[uuid4()] = CheckerIssue(
                        RuleCategory.FILENAME_CONSTRAINT,
                        f"File '{filename}' does not respect constraint 'contains {contains_constraint}'."
                    )       

    def check(self, hcl_folder_path: str):
        
        self._check_files_presence(hcl_folder_path)
        self._check_filenames_nomenclature(hcl_folder_path)

        for id, issue in self.report.items():
            print(issue)



from enum import Enum

from colorama import Style

from terracheck.types import RuleCategory

class CheckerIssue():

    def __init__(self, checking_category: RuleCategory.RuleCategory, message: str) -> None:
        self.checking_category     = checking_category.value
        self.message               = message

    def __str__(self) -> str:
        return f"[{self.checking_category[1]}{self.checking_category[0]}{Style.RESET_ALL}]: {self.message}"


import logging
import json
from uuid import uuid4

from terracheck.types.CheckerElement import CheckerElement
from terracheck.types.TerraformObjectType import TerraformObjectType

class HclCheckerReader():

    def __init__(self) -> None:
        self.extract      = {}
        self.checker_elements_ids = {
            TerraformObjectType.RESOURCE : [],
            TerraformObjectType.DATA : [],
            TerraformObjectType.VARIABLE : [],
            TerraformObjectType.OUTPUT : [],
            TerraformObjectType.TERRAFORM: [],
            TerraformObjectType.PROVIDER: []
        }

    @staticmethod
    def _get_tf_files_paths(folder: str):
        import os
        from os.path import isfile, join
        import glob

        # List .tf files in the folder
        tf_globs = glob.glob(join(folder,"*.tf"))
        # Keep files only
        tf_files = [f for f in tf_globs if isfile(f)]
        return tf_files

    @staticmethod
    def _get_hcl_content(file_path):
        import hcl2

        with open(file_path) as pfile:
            hcl_file = hcl2.load(pfile)
            return hcl_file

    def _read_resource_data(self, hcl_content: dict, terraform_object_type: TerraformObjectType, file_path: str):
        for terraform_resource in hcl_content:
            # print(json.dumps(hcl_content, indent=4))
            terraform_resource_type = list(terraform_resource.keys())[0]
            terraform_resource_given_name = list(terraform_resource[terraform_resource_type].keys())[0]
            checker_element = CheckerElement(
                                terraform_resource_given_name, 
                                file_path, 
                                terraform_object_type,
                                terraform_resource_type
                                )
            unique_id = uuid4()
            self.extract[unique_id] = checker_element
            self.checker_elements_ids[terraform_object_type].append(unique_id)

    def _read_variable_output(self, hcl_content: dict, terraform_object_type: TerraformObjectType, file_path: str):
        for terraform_resource in hcl_content:
            terraform_resource_type = list(terraform_resource.keys())[0]
            # next_key = list(terraform_resource[terraform_resource_type].keys())[0]
            additional_info = {}
            if terraform_object_type == TerraformObjectType.OUTPUT:
                additional_info["output_value"] = terraform_resource[terraform_resource_type]["value"]
            checker_element = CheckerElement(
                                terraform_resource_type, 
                                file_path, 
                                terraform_object_type,
                                None,
                                additional_info
                                )
            unique_id = uuid4()
            self.extract[unique_id] = checker_element
            self.checker_elements_ids[terraform_object_type].append(unique_id)

    def _read_terraform(self, hcl_content: dict, terraform_object_type: TerraformObjectType, file_path: str):
        for provider_alias_setup in hcl_content:
            required_providers = provider_alias_setup.get(TerraformObjectType.REQUIRED_PROVIDER.value, {})
            for alias_dict in required_providers:
                for alias in alias_dict:
                    checker_element = CheckerElement(
                                alias, 
                                file_path, 
                                TerraformObjectType.REQUIRED_PROVIDER,
                                None,
                                {}
                                )
                    unique_id = uuid4()
                    self.extract[unique_id] = checker_element
                    self.checker_elements_ids[terraform_object_type].append(unique_id)

    def _read_provider(self, hcl_content: dict, terraform_object_type: TerraformObjectType, file_path: str):
        for provider_setup in hcl_content:
            provider_setup_name = list(provider_setup.keys())[0]
            checker_element = CheckerElement(
                                provider_setup_name, 
                                file_path, 
                                terraform_object_type,
                                None,
                                {}
                                )
            unique_id = uuid4()
            self.extract[unique_id] = checker_element
            self.checker_elements_ids[terraform_object_type].append(unique_id)

    def read(self, folder_path):
        tf_files_paths = self._get_tf_files_paths(folder_path)

        for file in tf_files_paths:
            hcl_file_content = self._get_hcl_content(file)
            for str_terraform_object_type in hcl_file_content.keys():
                if not str_terraform_object_type in TerraformObjectType:
                    logging.warning(f"{str_terraform_object_type} is not part of terracheck.types.TerraformObjectType enumeration. Not considered.")
                else:
                    terraform_object_type = TerraformObjectType(str_terraform_object_type)
                    if terraform_object_type in [TerraformObjectType.RESOURCE, TerraformObjectType.DATA]:
                        self._read_resource_data(hcl_file_content[terraform_object_type.value], terraform_object_type, file)
                    elif terraform_object_type in [TerraformObjectType.VARIABLE, TerraformObjectType.OUTPUT]:
                        self._read_variable_output(hcl_file_content[terraform_object_type.value], terraform_object_type, file)
                    elif terraform_object_type == TerraformObjectType.TERRAFORM:
                        self._read_terraform(hcl_file_content[terraform_object_type.value], terraform_object_type, file)
                    elif terraform_object_type == TerraformObjectType.PROVIDER:
                        self._read_provider(hcl_file_content[terraform_object_type.value], terraform_object_type, file)
                        
# hcl_checker_reader = HclCheckerReader()
# extract = hcl_checker_reader.read("/home/wattache/projects/terraform_checker/tests/hcl_sample_1")

# for element_type, checker_element in extract.items():
#     print(json.dumps(checker_element.to_dict(), indent=4))


from enum import Enum

from colorama import Fore

class RuleCategory(Enum):
    NAME_HAS_DASH       = ("DASH_IN_NAME", Fore.LIGHTMAGENTA_EX)
    NAME_IS_RESOURCE    = ("NAME_IS_RESOURCE", Fore.CYAN)
    NAME_UNIQUENESS     = ("NAME_UNIQUENESS", Fore.BLUE)
    OUTPUT_EXPOSURE     = ("OUTPUT_EXPOSURE", Fore.YELLOW)
    FILE_PURPOSE        = ("FILE_PURPOSE", Fore.GREEN)
    MISSING_FILE        = ("MISSING_FILE", Fore.RED)
    PROVIDER_DEFINITION = ("PROVIDER_DEFINITION", Fore.LIGHTRED_EX)
    PROVIDER_ALIAS      = ("PROVIDER_ALIAS", Fore.LIGHTBLUE_EX)
    FILENAME_CONSTRAINT = ("FILENAME_CONSTRAINT", Fore.LIGHTGREEN_EX)


from typing import Dict
from uuid import uuid4
from terracheck.types.TerraformObjectType import TerraformObjectType
from terracheck.types.TerraformBasicFileName import TerraformBasicFileName
from terracheck.types.RuleCategory import RuleCategory
from terracheck.types.CheckerElement import CheckerElement
from terracheck.types.CheckerIssue import CheckerIssue

class StyleChecker():

    def __init__(self) -> None:
        self.report = {}
        self.metadata = {}
        self._init_metadata()

    def _init_metadata(self):
        self.metadata["checked_objects_total_count"] = 0
        self.metadata["dash_violations_count"] = 0
        self.metadata["resource_name_violations_count"] = 0
        self.metadata["output_location_count"] = 0
        self.metadata["output_exposure_count"] = 0
        self.metadata["resource_main_name_count"] = 0

    @staticmethod
    def _map_tf_object_type_to_file(terraform_object_type: TerraformObjectType) -> TerraformBasicFileName:
        if terraform_object_type.value == TerraformObjectType.OUTPUT.value:
            return TerraformBasicFileName.OUTPUTS_TF
        elif terraform_object_type.value == TerraformObjectType.VARIABLE.value:
            return TerraformBasicFileName.VARIABLES_TF
        elif terraform_object_type.value == TerraformObjectType.PROVIDER.value:
            return TerraformBasicFileName.VERSION_TF
        elif terraform_object_type.value == TerraformObjectType.REQUIRED_PROVIDER.value:
            return TerraformBasicFileName.VERSION_TF   
        else:
            return None 

    @staticmethod
    def _group_extract_by_unique_terraform_resource_type(hcl_reader_extract: Dict[str, CheckerElement]):
        resources_clusters = {}
        not_unique_resource_types = []
        for _, element in hcl_reader_extract.items():
            ert = element.terraform_resource_type
            if ert is not None:
                if ert in resources_clusters:
                    del resources_clusters[ert]
                    not_unique_resource_types.append(ert)
                elif ert in not_unique_resource_types:
                    pass
                else:
                    resources_clusters[ert] = element
                
        return resources_clusters

    def _check_no_dash_only_underscores_in_objects_names(self, hcl_reader_extract: Dict[str, CheckerElement]):
        for _, element in hcl_reader_extract.items():
            if element.terraform_object_type not in [TerraformObjectType.PROVIDER, TerraformObjectType.REQUIRED_PROVIDER]:
                dash_count = element.name.count('-')
                if dash_count >= 1:
                    resource_type_display = f"'{element.terraform_resource_type}'" if element.terraform_resource_type is not None else "named"
                    self.report[uuid4()] = CheckerIssue(
                        RuleCategory.NAME_HAS_DASH, 
                        f"Name of {element.terraform_object_type.value} {resource_type_display} '{element.name}' contains dash ('-'). Shoud be underscores ('_')."
                    )
                    self.metadata["checked_objects_total_count"] += 1
                    self.metadata["dash_violations_count"] += 1

    def _check_resource_name_is_not_resource_type(self, hcl_reader_extract: Dict[str, CheckerElement]):
        for _, element in hcl_reader_extract.items():
            if element.terraform_resource_type is not None and element.name == element.terraform_resource_type:
                self.report[uuid4()] = CheckerIssue(
                    RuleCategory.NAME_IS_RESOURCE, 
                    f"{element.terraform_object_type.value.capitalize()} '{element.terraform_resource_type}' is named '{element.name}'. Chose a name different than the resource name."
                )
                self.metadata["checked_objects_total_count"] += 1
                self.metadata["resource_name_violations_count"] += 1

    def _check_output_variable_location(self, hcl_reader_extract: Dict[str, CheckerElement]):
        for _, element in hcl_reader_extract.items():
            expected_file = self._map_tf_object_type_to_file(element.terraform_object_type)
            if expected_file is not None and element.filename != expected_file.value:
                self.report[uuid4()] = CheckerIssue(
                    RuleCategory.FILE_PURPOSE,
                    f"{element.terraform_object_type.value.capitalize()} '{element.name}' should be declared in file {expected_file.value} not in {element.filename}."
                )
                self.metadata["checked_objects_total_count"] += 1
                self.metadata["output_location_count"] += 1

    def _check_output_exposure(self, hcl_reader_extract: Dict[str, CheckerElement]):
        for _, element in hcl_reader_extract.items():
            if element.terraform_object_type.value == TerraformObjectType.OUTPUT.value and "var." in element.additional_info["output_value"]:
                self.report[uuid4()] = CheckerIssue(
                    RuleCategory.OUTPUT_EXPOSURE,
                    f"Output '{element.name}' references a variable as value. It should not."
                )
                self.metadata["checked_objects_total_count"] += 1
                self.metadata["output_exposure_count"] += 1

    def _check_unique_resources_are_named_main(self, hcl_reader_extract: Dict[str, CheckerElement]):
        unique_resources = self._group_extract_by_unique_terraform_resource_type(hcl_reader_extract)
        for element in unique_resources.values():
            self.report[uuid4()] = CheckerIssue(
                    RuleCategory.NAME_UNIQUENESS,
                    f"{element.terraform_object_type.value.capitalize()} '{element.terraform_resource_type}' is unique, and thus should be named 'main', not '{element.name}'."
                )
            self.metadata["checked_objects_total_count"] += 1
            self.metadata["resource_main_name_count"] += 1


    def check(self, hcl_reader_extract: Dict[str, CheckerElement]):
        
        self._check_no_dash_only_underscores_in_objects_names(hcl_reader_extract)
        self._check_resource_name_is_not_resource_type(hcl_reader_extract)
        self._check_output_variable_location(hcl_reader_extract)
        self._check_output_exposure(hcl_reader_extract)
        self._check_unique_resources_are_named_main(hcl_reader_extract)

        for id, issue in self.report.items():
            print(issue)



from enum import Enum, EnumMeta

class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True    


class BaseEnum(Enum, metaclass=MetaEnum):
    pass

class TerraformObjectType(BaseEnum):
    OUTPUT            = "output"
    VARIABLE          = "variable"
    RESOURCE          = "resource"
    DATA              = "data"
    TERRAFORM         = "terraform"
    PROVIDER          = "provider"
    REQUIRED_PROVIDER = "required_providers"


#!/usr/bin/env python

from asyncore import read
from setuptools import setup, find_packages

# Reading long description from file
with open('README.md') as file:
    long_description = file.read()

# Specify requirements of your package here
REQUIREMENTS = [
    'python-hcl2==3.0.5',
    'colorama==0.4.5',
    'pytest==7.2.0',
    'PyYAML==6.0'
  ]

# Some more details
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Utilities',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    ]

setup(name='terracheck',
      version='0.1.1',
      description='Python Terraform checker from GCP good practices',
      long_description=long_description,
      long_description_content_type="text/markdown",
      #url='https://github.com/nikhilkumarsingh/mygmap',
      author='William Attache',
      author_email='william.attache@mydevacademy.com',
      license='MIT',
      packages=find_packages(exclude=('tests', 'docs')),
      classifiers=CLASSIFIERS,
      install_requires=REQUIREMENTS,
      keywords='terraform gcp goodpractices'
     )